// Nathan E. Moore
// March 2, 2005

#include "windows_include.h"
#include "checkedassign.h"
#include "storage.h"
#include "menustorage.h"
#include "taskalloc.h"
#include "idlist.h"
#include "comrefcounted.h"
#include <sstream>
#include <lsapi/lsapi.h>
#include <loki/SmartPtr.h>
#include <loki/Singleton.h>
#include <atlbase.h>
#include <atlwin.h>
#include <shlobj.h>
#include <strsafe.h>

namespace contextmenu
{
	class ContextMenuWindow;
	typedef Loki::SingletonHolder<ContextMenuWindow> CMWnd;
	const char* BANG_ID = "!ContextMenu";
	const char* MODULE_NAME = "ContextMenu";

	typedef Loki::SmartPtr<wchar_t, Loki::RefCounted, Loki::DisallowConversion, Loki::AssertCheck, utility::ArrayStorage> SmartWStrBuf;
	SmartWStrBuf ToUnicode(const std::string&);

	class ContextMenuWindow :
		public ATL::CWindowImpl<ContextMenuWindow, ATL::CWindow, ATL::CWinTraits<WS_POPUP, WS_EX_TOOLWINDOW | WS_EX_NOPARENTNOTIFY> >
	{
	public:
		DECLARE_WND_CLASS(MODULE_NAME);

		BEGIN_MSG_MAP(ContextMenuWindow)
			MESSAGE_HANDLER(WM_INITMENUPOPUP, OnICM2)
			MESSAGE_HANDLER(WM_DRAWITEM, OnICM2)
			MESSAGE_HANDLER(WM_MEASUREITEM, OnICM2)			
			MESSAGE_HANDLER(WM_MENUCHAR, OnICM3)
		END_MSG_MAP()

		void PopupContextMenu(const std::string& path, int x, int y, DWORD flags);

	private:
		ContextMenuWindow()
			:icm2_(),
			icm3_()
		{
			// this is only really needed if contextmenu is run as a threaded module
			CoInitialize(0);

			Create(0);
		}

		~ContextMenuWindow()
		{
			if (IsWindow())
				DestroyWindow();
		
			CoUninitialize();
		}

		LRESULT OnICM2(UINT msg, WPARAM wp, LPARAM lp, BOOL& handled)
		{
			if (icm3_)
				icm3_->HandleMenuMsg(msg, wp, lp);
			else if (icm2_)
				icm2_->HandleMenuMsg(msg, wp, lp);
			else
				handled = FALSE;
			return 0;
		}

		
		LRESULT OnICM3(UINT msg, WPARAM wp, LPARAM lp, BOOL& handled)
		{
			if (icm3_)
			{
				LRESULT lres;
				icm3_->HandleMenuMsg2(msg, wp, lp, &lres);
				return lres;
			}
			else
			{
				handled = FALSE;
				return 0;
			}
		}
		
		friend struct Loki::CreateUsingNew<ContextMenuWindow>;

		// no copies
        ContextMenuWindow(const ContextMenuWindow&);
		ContextMenuWindow& operator=(const ContextMenuWindow&);
		
		Loki::SmartPtr<IContextMenu2, com::COMRefCounted> icm2_;
		Loki::SmartPtr<IContextMenu3, com::COMRefCounted> icm3_;
	};	// end class ContextMenu

	void ContextMenuWindow::PopupContextMenu(const std::string& path, int x, int y, DWORD flags)
	{
		using std::runtime_error;
		typedef Loki::SmartPtr<ITEMIDLIST, Loki::NoCopy, Loki::AllowConversion, Loki::AssertCheck, com::TaskAllocStorage> IdListPtr;
		
		Loki::SmartPtr<IShellFolder, com::COMRefCounted> desktop;
		if (FAILED(CoCreateInstance(CLSID_ShellDesktop, NULL, CLSCTX_INPROC_SERVER | CLSCTX_INPROC_HANDLER, 
			IID_IShellFolder, reinterpret_cast<void**>(&GetImplRef(desktop)))))
			throw runtime_error(ERROR_MSG("Could not get desktop IShellFolder"));

		SmartWStrBuf ustr(ToUnicode(path));

		IdListPtr idlist;
		if (FAILED(desktop->ParseDisplayName(*this, NULL, GetImpl(ustr), NULL, &(GetImplRef(idlist)), 0)))
			throw runtime_error(ERROR_MSG("Could not create a ITEMIDLIST from given path"));

		IdListPtr parent(com::idlist::ContainingElement(idlist));
		Loki::SmartPtr<IShellFolder, com::COMRefCounted> folder;
		if (com::idlist::IsDesktop(parent))
			folder = desktop;
		else if (FAILED(desktop->BindToObject(parent, NULL, IID_IShellFolder, reinterpret_cast<void**>(&(GetImplRef(folder))))))
			throw runtime_error(ERROR_MSG("Could not create IShellfolder for given path"));

		IdListPtr file(com::idlist::LastElement(idlist));
		Loki::SmartPtr<IContextMenu, com::COMRefCounted> cmenu;
		if (FAILED(folder->GetUIObjectOf(*this, 1, const_cast<const ITEMIDLIST**>(&(GetImplRef(file))), IID_IContextMenu, 0, reinterpret_cast<void**>(&(GetImplRef(cmenu))))))
			throw runtime_error(ERROR_MSG("Could not get IContextMenu for given path"));

		utility::SmartMenu menu(CreatePopupMenu());
		if (!menu)
			throw runtime_error(ERROR_MSG("Could not create a menu"));
		if (FAILED(cmenu->QueryContextMenu(menu, 0, 1, 0x7FFF, CMF_NORMAL)))
			throw runtime_error(ERROR_MSG("QueryContextMenu failed")); 

		Loki::SmartPtr<IContextMenu3, com::COMRefCounted> cm3;
		Loki::SmartPtr<IContextMenu2, com::COMRefCounted> cm2;
		if (SUCCEEDED(cmenu->QueryInterface(IID_IContextMenu3, reinterpret_cast<void**>(&GetImplRef(cm3)))))
			icm3_ = cm3;
		else if (SUCCEEDED(cmenu->QueryInterface(IID_IContextMenu2, reinterpret_cast<void**>(&GetImplRef(cm3)))))
			icm2_ = cm2;	

		int cmd = TrackPopupMenu(menu, TPM_RECURSE | flags | TPM_LEFTBUTTON | TPM_RETURNCMD , x, y, 0, *this, NULL);
				
		icm3_ = NULL;
		icm2_ = NULL;

		if (0 != cmd)
		{
			CMINVOKECOMMANDINFOEX cmdinfo;
			memset(&cmdinfo, 0, sizeof(cmdinfo));
			cmdinfo.cbSize = sizeof(cmdinfo);
			cmdinfo.fMask = CMIC_MASK_FLAG_NO_UI | CMIC_MASK_UNICODE;
			cmdinfo.hwnd = *this;
			cmdinfo.lpVerb = MAKEINTRESOURCEA(cmd - 1);
			cmdinfo.lpVerbW = MAKEINTRESOURCEW(cmd - 1);
			cmdinfo.nShow = SW_SHOW;

			if (FAILED(cmenu->InvokeCommand(reinterpret_cast<CMINVOKECOMMANDINFO*>(&cmdinfo))))
				throw runtime_error(ERROR_MSG("InvokeCommand Failed"));
		}
	}

	SmartWStrBuf ToUnicode(const std::string& str)
	{
		// double the character count of the buffer so if the given string
		// converts to all sequence points the buffer is still large enough
		int wbuflen(str.size() * 2 + 1);

		SmartWStrBuf wbuffer(new wchar_t[wbuflen]);

		// use c_str and -1 instead of data and size so that MB2WC will null terminate the string for us.
		if (0 < MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, str.c_str(), -1, GetImpl(wbuffer), wbuflen))
			return wbuffer;
		else
		{
			// the huge buffer allocated above should prevent this error from ever occuring
			// if it does there is an error in the above code.
			ASSERTION(ERROR_INSUFFICIENT_BUFFER != GetLastError());
			throw std::runtime_error(ERROR_MSG("Error during conversion to Unicode."));
		}
	}

	void ContextMenuBang(HWND, LPCSTR args)
	{
		try
		{
			enum { MAX_TOKEN = 4096 };
			char token[MAX_TOKEN];
			std::string path;

			if (GetToken(args, token, &args, FALSE))
				path = token;
			else
			{
				LSLog(LOG_ERROR, MODULE_NAME, "Malformed bang arguments: No path given");
				return;	// path is required
			}

			utility::CheckedAssign<int> x, y;
			utility::CheckedAssign<UINT> horz(TPM_LEFTALIGN);
			utility::CheckedAssign<UINT> vert(TPM_TOPALIGN);

			while (GetToken(args, token, &args, FALSE) 
				&& !(x.IsAssigned() && y.IsAssigned() && horz.IsAssigned() && vert.IsAssigned()))
			{
				if ('0' <= *token && '9' >= *token || '-' == *token || '+' == *token)
				{
					if (!x.IsAssigned())
						x = ParseCoordinate(token, 0, LSGetSystemMetrics(SM_CXSCREEN));
					else if (!y.IsAssigned())
						y = ParseCoordinate(token, 0, LSGetSystemMetrics(SM_CYSCREEN));
				}
				else if (0 == _strnicmp(token, "x=", 2))
					x = ParseCoordinate(token + 2, 0, LSGetSystemMetrics(SM_CXSCREEN));
				else if (0 == _strnicmp(token, "y=", 2))
					y = ParseCoordinate(token + 2, 0, LSGetSystemMetrics(SM_CYSCREEN));
				else if (0 == _stricmp(token, "left"))
					horz = TPM_LEFTALIGN;
				else if (0 == _stricmp(token, "right"))
					horz = TPM_RIGHTALIGN;
				else if (0 == _stricmp(token, "top"))
					vert = TPM_TOPALIGN;
				else if (0 == _stricmp(token, "bottom"))
					vert = TPM_BOTTOMALIGN;
				else if (0 == _stricmp(token, "center")) // catch all center
				{
					if (!horz.IsAssigned())
						horz = TPM_CENTERALIGN;
					if (!vert.IsAssigned())
						vert = TPM_VCENTERALIGN;
				}
			}

			if (!(x.IsAssigned() && y.IsAssigned()))
			{	
				POINT pt;
				GetCursorPos(&pt);
				// POINT is define as LONG values however TrackPopupMenu uses ints
				x = static_cast<int>(pt.x);
				y = static_cast<int>(pt.y);
			}

			CMWnd::Instance().PopupContextMenu(path, x.GetValue(), y.GetValue(), horz.GetValue() | vert.GetValue());
		}
		catch (std::exception& e)
		{
			std::ostringstream oss;
			// e.what() should include location info
			oss << BANG_ID << "failed: " << e.what();
			LSLog(LOG_ERROR, MODULE_NAME, oss.str().c_str());
		}
		catch (...)
		{
			std::ostringstream oss;
			oss << BANG_ID << "failed.";
			LSLog(LOG_ERROR, MODULE_NAME, oss.str().c_str());
		}
	}
}	// end namespace contextmenu

 // Litestep module interface (no prototypes) 
extern "C" __declspec(dllexport) int initModuleEx(HWND, HINSTANCE, LPCSTR)
{
	if (AddBangCommand(contextmenu::BANG_ID, &contextmenu::ContextMenuBang))
		return 0;
	else
		return 1;
}

extern "C" __declspec(dllexport) void quitModule(HINSTANCE)
{
	// don't check this return value
	// removal should be a no fail op much like std::free
	// since AddBangCommand will never fail
	RemoveBangCommand(contextmenu::BANG_ID);
}
